using System;
using System.Linq;
using System.Web;
using Microsoft.WindowsAzure;
using Microsoft.WindowsAzure.Diagnostics;
using Microsoft.WindowsAzure.ServiceRuntime;
using Microsoft.WindowsAzure.StorageClient;



namespace CloudUI
{
    public class WebRole : RoleEntryPoint
    {
        private CloudBlobClient _blob_cli;
        private const int PAGE_SIZE = 512;

        public override bool OnStart()
        {            
            CloudStorageAccount.SetConfigurationSettingPublisher((configName, configSetter) =>
            {
              configSetter(RoleEnvironment.GetConfigurationSettingValue(configName));

              RoleEnvironment.Changed += (sender, arg) =>
              {
                if (arg.Changes.OfType<RoleEnvironmentConfigurationSettingChange>()
                  .Any((change) => (change.ConfigurationSettingName == configName)))
                {
                  if (!configSetter(RoleEnvironment.GetConfigurationSettingValue(configName)))
                  {
                    RoleEnvironment.RequestRecycle();
                  }
                }
              };
            });

            DiagnosticMonitorConfiguration config = DiagnosticMonitor.GetDefaultInitialConfiguration();
            config.Logs.ScheduledTransferPeriod = TimeSpan.FromMinutes(1);
            config.Logs.ScheduledTransferLogLevelFilter = LogLevel.Error;
            config.Logs.BufferQuotaInMB = 100;

            config.DiagnosticInfrastructureLogs.ScheduledTransferPeriod =
                TimeSpan.FromMinutes(1);
            config.DiagnosticInfrastructureLogs.ScheduledTransferLogLevelFilter =
                LogLevel.Critical;
            config.DiagnosticInfrastructureLogs.BufferQuotaInMB = 100;

            config.Directories.ScheduledTransferPeriod
                = System.TimeSpan.FromMinutes(1);
            config.Directories.BufferQuotaInMB = 200;

            config.WindowsEventLog.DataSources.Add("Application!*");
            config.WindowsEventLog.DataSources.Add("System!*");
            config.WindowsEventLog.ScheduledTransferPeriod
              = TimeSpan.FromMinutes(1);
            config.WindowsEventLog.ScheduledTransferLogLevelFilter
              = LogLevel.Warning;
            config.WindowsEventLog.BufferQuotaInMB = 100;

            config.PerformanceCounters.DataSources.Add(
              new PerformanceCounterConfiguration()
              {
                  CounterSpecifier = @"\Processor(*)\% Processor Time",
                  SampleRate = TimeSpan.FromSeconds(60)
              });
            config.PerformanceCounters.ScheduledTransferPeriod = TimeSpan.FromMinutes(1);
            config.PerformanceCounters.BufferQuotaInMB = 100;

            LocalResource mylog =
             RoleEnvironment.GetLocalResource("MyLog");
            config.Directories.DataSources.Add(
              new DirectoryConfiguration()
              {
                  Container = "wad-custom-log",
                  Path = mylog.RootPath,
                  DirectoryQuotaInMB = 10
              }
            );
            config.Directories.ScheduledTransferPeriod =
              System.TimeSpan.FromMinutes(1);

            CrashDumps.EnableCollection(true);
            config.Directories.ScheduledTransferPeriod =
              System.TimeSpan.FromMinutes(1);


            DiagnosticMonitor.Start("DiagnosticsConnectionString", config);
            // For information on handling configuration changes
            // see the MSDN topic at http://go.microsoft.com/fwlink/?LinkId=166357.
            //RoleEnvironment.Changing += RoleEnvironmentChanging;

            return base.OnStart();
        }

        private void RoleEnvironmentChanging(object sender, RoleEnvironmentChangingEventArgs e)
        {
            // If a configuration setting is changing
            if (e.Changes.Any(change => change is RoleEnvironmentConfigurationSettingChange))
            {
                // Set e.Cancel to true to restart this role instance
                e.Cancel = true;
            }
        }

        public override void Run()
        {
            var account = CloudStorageAccount.FromConfigurationSetting("MyStorage");
            var table_cli = account.CreateCloudTableClient();
            table_cli.RetryPolicy = RetryPolicies.Retry(3, TimeSpan.FromMilliseconds(500));
            table_cli.CreateTableIfNotExist("Message"); 
            _blob_cli = account.CreateCloudBlobClient();
            _blob_cli.RetryPolicy = RetryPolicies.Retry(3, TimeSpan.FromMilliseconds(500));
            this.CreateContainer("img0general", "");
            this.CreateContainer("img1family", "Ƒ");
            this.CreateContainer("img2view", "i");
            this.CreateContainer("queorg", "IWi摜");
            this.CreateContainer("quethumb", "TlC摜");
            var que_cli = account.CreateCloudQueueClient();
            var que1 = que_cli.GetQueueReference("thumbnails");
            que1.CreateIfNotExist();
            var que2 = que_cli.GetQueueReference("delimages");
            que2.CreateIfNotExist();
            var c = this.CreateContainer("images", "fBNg");
            var pageBlob = c.GetPageBlobReference("pages");
            if (!this.isExists(pageBlob))
            {
              pageBlob.Create(PAGE_SIZE * 1024 * 1024);
            }

            var dc = this.CreateContainer("drives", "hCu");
            var p_drive = dc.GetPageBlobReference("mydrive");
            if (!this.isExists(p_drive))
            {
                p_drive.Create(50 * 1024 * 1024);
                LocalResource localCache =
                    RoleEnvironment.GetLocalResource("MyCache");
                CloudDrive.InitializeCache(localCache.RootPath,
                    localCache.MaximumSizeInMegabytes);
                var d = new CloudDrive(p_drive.Uri, account.Credentials);
                //d.Delete();
                d.Create(50);
            }
            base.Run();
        }

        private CloudBlobContainer CreateContainer(string name, string dispname)
        {
            var container = _blob_cli.GetContainerReference(name);
            container.CreateIfNotExist();
            container.Metadata.Add("Name", HttpUtility.UrlEncode(dispname));
            container.SetMetadata();
            var perms = container.GetPermissions();
            perms.PublicAccess = BlobContainerPublicAccessType.Container;
            container.SetPermissions(perms);
            return container;
        }

        private bool isExists(CloudBlob blob)	
        {
          try
          {
            blob.FetchAttributes();
            return true;
          }
          catch(StorageClientException e) {
            if (e.ErrorCode == StorageErrorCode.ResourceNotFound)
            {
              return false;
            }
            else
            {
              throw;
            }
          }
        }	
    }
}
