﻿Imports Microsoft.WindowsAzure
Imports Microsoft.WindowsAzure.Diagnostics
Imports Microsoft.WindowsAzure.ServiceRuntime
Imports Microsoft.WindowsAzure.StorageClient

Public Class WebRole
    Inherits RoleEntryPoint

    Private _blob_cli As CloudBlobClient
    Private Const PAGE_SIZE As Integer = 512

    Public Overrides Function OnStart() As Boolean
        CloudStorageAccount.SetConfigurationSettingPublisher(Function(configName, configSetter) SetConfig(configName, configSetter))
        Dim config = DiagnosticMonitor.GetDefaultInitialConfiguration()

        config.Logs.ScheduledTransferPeriod = TimeSpan.FromMinutes(1)
        config.Logs.ScheduledTransferLogLevelFilter = LogLevel.Error
        config.Logs.BufferQuotaInMB = 100

        config.DiagnosticInfrastructureLogs.ScheduledTransferPeriod =
            TimeSpan.FromMinutes(1)
        config.DiagnosticInfrastructureLogs.ScheduledTransferLogLevelFilter =
            LogLevel.Critical
        config.DiagnosticInfrastructureLogs.BufferQuotaInMB = 100

        config.Directories.ScheduledTransferPeriod =
            System.TimeSpan.FromMinutes(1)
        config.Directories.BufferQuotaInMB = 100

        config.WindowsEventLog.DataSources.Add("Application!*")
        config.WindowsEventLog.DataSources.Add("System!*")
        config.WindowsEventLog.ScheduledTransferPeriod = TimeSpan.FromMinutes(1)
        config.WindowsEventLog.ScheduledTransferLogLevelFilter = LogLevel.Warning
        config.WindowsEventLog.BufferQuotaInMB = 100

        config.PerformanceCounters.DataSources.Add(
            New PerformanceCounterConfiguration() With
            {
                .CounterSpecifier = "\Processor(*)\% Processor Time",
                .SampleRate = TimeSpan.FromSeconds(60)
            })
        config.PerformanceCounters.ScheduledTransferPeriod = TimeSpan.FromMinutes(1)
        config.PerformanceCounters.BufferQuotaInMB = 100

        CrashDumps.EnableCollection(True)

        Dim mylog = RoleEnvironment.GetLocalResource("MyLog")
        config.Directories.DataSources.Add(
                New DirectoryConfiguration() With
                {
                    .Container = "wad-custom-log",
                    .DirectoryQuotaInMB = 10,
                    .Path = mylog.RootPath
                }
        )
        config.Directories.ScheduledTransferPeriod = TimeSpan.FromMinutes(1)

        DiagnosticMonitor.Start("DiagnosticsConnectionString", config)
        ' For information on handling configuration changes
        ' see the MSDN topic at http://go.microsoft.com/fwlink/?LinkId=166357.
        'AddHandler RoleEnvironment.Changing, AddressOf RoleEnvironmentChanging

        Return MyBase.OnStart()
    End Function

    Private Function SetConfig(ByVal configName As String, ByVal configSetter As System.Func(Of String, Boolean)) As Object
        configSetter(RoleEnvironment.GetConfigurationSettingValue(configName))
        AddHandler RoleEnvironment.Changed, Function(sender, arg) SetConfigSetter(sender, arg, configName, configSetter)
        Return Nothing
    End Function

    Private Function SetConfigSetter(ByVal sender As Object, ByVal arg As Microsoft.WindowsAzure.ServiceRuntime.RoleEnvironmentChangedEventArgs, ByVal configName As String, ByVal configSetter As System.Func(Of String, Boolean)) As Object
        If arg.Changes.OfType(Of RoleEnvironmentConfigurationSettingChange)().Any(Function(change) (change.ConfigurationSettingName Is configName)) Then
            If Not configSetter(RoleEnvironment.GetConfigurationSettingValue(configName)) Then
                RoleEnvironment.RequestRecycle()
            End If
        End If
        Return Nothing
    End Function

    Private Sub RoleEnvironmentChanging(ByVal sender As Object, ByVal e As RoleEnvironmentChangingEventArgs)

        ' If a configuration setting is changing
        If (e.Changes.Any(Function(change) TypeOf change Is RoleEnvironmentConfigurationSettingChange)) Then
            ' Set e.Cancel to true to restart this role instance
            e.Cancel = True
        End If

    End Sub

    Public Overrides Sub Run()
        Dim account = CloudStorageAccount.FromConfigurationSetting("MyStorage")
        Dim table_cli = account.CreateCloudTableClient()
        table_cli.RetryPolicy = RetryPolicies.Retry(3, TimeSpan.FromMilliseconds(500))
        table_cli.CreateTableIfNotExist("Message")
        _blob_cli = account.CreateCloudBlobClient()
        _blob_cli.RetryPolicy = RetryPolicies.Retry(3, TimeSpan.FromMilliseconds(500))

        Me.CreateContainer("img0general", "一般")
        Me.CreateContainer("img1family", "家族")
        Me.CreateContainer("img2view", "風景")
        Me.CreateContainer("queorg", "オリジナル画像")
        Me.CreateContainer("quethumb", "サムネイル画像")

        Dim que_cli = account.CreateCloudQueueClient()
        Dim que1 = que_cli.GetQueueReference("thumbnails")
        que1.CreateIfNotExist()
        Dim que2 = que_cli.GetQueueReference("delimages")
        que2.CreateIfNotExist()

        Dim c = Me.CreateContainer("images", "ディレクトリ")
        Dim pageBlob = c.GetPageBlobReference("pages")
        If Not Me.isExists(pageBlob) Then
            pageBlob.Create(PAGE_SIZE * 1024 * 1024)
        End If

        Dim dc = Me.CreateContainer("drives", "ドライブ")
        Dim p_drive = dc.GetPageBlobReference("mydrive")
        If Not Me.isExists(p_drive) Then
            p_drive.Create(50 * 1024 * 1024)
            Dim localCache =
                RoleEnvironment.GetLocalResource("MyCache")
            CloudDrive.InitializeCache(localCache.RootPath,
                localCache.MaximumSizeInMegabytes)
            Dim d = New CloudDrive(p_drive.Uri, account.Credentials)
            'd.Delete()
            d.Create(50)
        End If
        MyBase.Run()
    End Sub

    Private Function CreateContainer(ByVal name As String,
        ByVal dispname As String) As CloudBlobContainer
        Dim container = _blob_cli.GetContainerReference(name)
        container.CreateIfNotExist()
        container.Metadata.Add("Name", HttpUtility.UrlEncode(dispname))
        container.SetMetadata()
        Dim perms = container.GetPermissions()
        perms.PublicAccess = BlobContainerPublicAccessType.Container
        container.SetPermissions(perms)
        Return container
    End Function

    Private Function isExists(ByVal blob As CloudBlob) As Boolean
        Try
            blob.FetchAttributes()
            Return True
        Catch e As StorageClientException
            If e.ErrorCode = StorageErrorCode.ResourceNotFound Then
                Return False
            Else
                Throw
            End If
        End Try
    End Function

End Class

